# -*- coding: utf-8 -*-
"""
Created on Fri Apr 5 16:22:00 2024

@author: vlandgraf
"""

# Import necessary libraries and modules
import os
import glob
import copy
import pickle as pkl
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from scipy import fftpack
from joblib import Parallel, delayed
from pymatgen.core import Lattice
from pymatgen.io.vasp import Vasprun
from pymatgen.io.vasp.outputs import Poscar
import gemdat
from gemdat import SitesData, Trajectory, plots, simulation_metrics
from gemdat.io import load_known_material
import vasppy.rdf
import general_AF_structure_analysis_special as general_AF_structure_analysis

# Load trajectory data from cache
#trajctory may also be loaded from xml directly Trajectory.from_vasprun('vasprun.xml')

trajectory = Trajectory.from_cache('vasprun.xml.69adc856.cache')

# Extract the lattice information from the trajectory
lattice = Lattice.from_dict(trajectory[0].lattice.as_dict())

def calculate_rdf_parallelized(center_coords, diff_coords, lattice, resolution=0.051, max_distance=10.0):
    """
    Calculate the Radial Distribution Function (RDF) in a parallelized manner.
    
    Args:
        center_coords (ndarray): Coordinates of the center atoms.
        diff_coords (ndarray): Coordinates of the surrounding atoms.
        lattice (Lattice): Lattice object representing the periodic boundary conditions.
        resolution (float): Bin size for the RDF histogram.
        max_distance (float): Maximum distance to consider for RDF.
        
    Returns:
        bins (ndarray): Bin edges for the RDF.
        rdf (ndarray): Calculated RDF values.
    """

    try:
        num_time_steps, num_atoms, num_dimensions = diff_coords.shape
    except ValueError:
        num_time_steps = 1
        num_atoms, num_dimensions = diff_coords.shape

    # Calculate the particle volume
    particle_vol = num_atoms / ((4 / 3) * np.pi * max_distance ** 3)

    # Function to calculate distances for each timestep
    def calculate_distances(t):
        return lattice.get_all_distances(center_coords[t, :, :], diff_coords[t, :, :])

    # Parallel computation of distances across time steps
    all_dists = Parallel(n_jobs=-1)(delayed(calculate_distances)(t) for t in range(num_time_steps))

    # Flatten and filter non-zero distances
    distances = np.concatenate([dists[dists != 0].flatten() for dists in all_dists])

    # Create bins for RDF calculation
    bins = np.arange(0, max_distance + resolution, resolution)

    # Compute histogram for RDF
    rdf, _ = np.histogram(distances, bins=bins, density=False)

    # Normalize RDF based on the volume of spherical shells
    norm = np.array([(4 / 3) * np.pi * ((r + resolution) ** 3 - r ** 3) * particle_vol for r in bins[:-1]])
    rdf = np.array([rdf[i] / norm[i] for i in range(len(rdf))])

    return bins, rdf

# Filter for Li atoms and calculate the RDF
bins, rdf = calculate_rdf_parallelized(trajectory.filter('Li').coords, trajectory.filter('Li').coords, lattice)

# Plot the RDF
plt.figure(figsize=(3.25, 3.25), dpi=400)
plt.plot(bins[:-1], rdf)
plt.tick_params(axis='both', direction='in')
plt.ylabel('Li-Li RDF (arbitrary units)')
plt.xlabel('Radius (Å)')
plt.yticks([])
plt.savefig('rdf.png', format='png', bbox_inches='tight')
