# -*- coding: utf-8 -*-
"""
Created on Fri Apr 5 07:13:16 2024

@author: vlandgraf

Playing with van Hove correlation functions
"""

# Import necessary libraries and modules
import os
import glob
import copy
import pickle as pkl
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from scipy import fftpack
from joblib import Parallel, delayed
from pymatgen.core import Lattice
from pymatgen.io.vasp import Vasprun
from pymatgen.io.vasp.outputs import Poscar
import gemdat
from gemdat import SitesData, Trajectory, plots, simulation_metrics
from gemdat.io import load_known_material
import vasppy.rdf
import general_AF_structure_analysis_special as general_AF_structure_analysis

# Load trajectory data from cache
trajectory = Trajectory.from_cache('vasprun.xml.69adc856.cache')

# Extract the lattice information from the trajectory
lattice = Lattice.from_dict(trajectory[0].lattice.as_dict())

def calculate_van_hove_with_time_skip(coords, lattice, resolution=0.1, max_distance=10.0, time_step_skip=50):
    """
    Calculate the van Hove correlation function with a time step skip.
    
    Args:
        coords (ndarray): Coordinates of the atoms.
        lattice (Lattice): Lattice object representing the periodic boundary conditions.
        resolution (float): Bin size for the histogram.
        max_distance (float): Maximum distance to consider for the van Hove function.
        time_step_skip (int): Number of time steps to skip for each calculation.
        
    Returns:
        bins (ndarray): Bin edges for the van Hove function.
        self_correlation_2D_array (ndarray): 2D array of van Hove correlation function over time.
    """

    # Determine the number of time steps, atoms, and dimensions
    num_time_steps_formally, num_atoms, num_dimensions = coords.shape

    # Calculate particle density
    particle_density = num_atoms / ((4 / 3) * np.pi * max_distance ** 3)

    # Define bins for the histogram
    bins = np.arange(0, max_distance + resolution, resolution)

    # Normalization based on the volume of spherical shells
    norm = np.array([(4 / 3) * np.pi * ((r + resolution) ** 3 - r ** 3) * particle_density for r in bins[:-1]])

    # Define the time steps to calculate
    times = np.arange(time_step_skip, num_time_steps_formally, time_step_skip)

    # Function to calculate distances at each time step
    def calculate_distances(t):
        distances = []
        for sub_time in np.arange(t - time_step_skip, t, 1):
            distances.extend([lattice.get_distance_and_image(i, j)[0] for i, j in zip(coords[sub_time, :], coords[0, :])])
        distances = np.array(distances)

        # Calculate histogram for distances
        rdf, _ = np.histogram(distances, bins=bins, density=False)

        # Normalize RDF by the number of atoms and time steps
        rdf = rdf / (len(coords[0]) * time_step_skip)
        return rdf

    # Parallel computation of the van Hove correlation function over time
    self_correlation_2D_array = np.array(
        Parallel(n_jobs=-1)(delayed(calculate_distances)(t) for t in times)
    )

    return bins[:-1], self_correlation_2D_array

# Filter for Li atoms and calculate the van Hove function
coords = trajectory.filter('Li').coords[::1]
bins, two2d_van_hove = calculate_van_hove_with_time_skip(coords, lattice, resolution=0.2, time_step_skip=50)

# Plot the 2D van Hove correlation function
fig, ax = plt.subplots(1, 1, figsize=(3.25, 3.25), dpi=400)
img = ax.imshow(two2d_van_hove, cmap='viridis', interpolation='nearest', aspect='auto', vmin=0, vmax=0.1)
cbar = fig.colorbar(img)
ax.set_xticklabels(ax.get_xticks() / 5, size=9)
ax.set_ylabel('Simulation time, t (ps)', size=10)
ax.set_xlabel('Radius, r (Å)', size=10)

# Adjust y-tick labels to represent time in picoseconds
y_tick_labels = [int(i * 50 * 1E-15 / 1E-12) for i in ax.get_yticks()]
ax.set_yticklabels(y_tick_labels, size=9)
cbar.set_label('normalized $G^{self}$ (r, t)', size=10)

# Save the plot
plt.savefig('2D_vanHove_self_correlation.png', format='png', dpi=400, bbox_inches='tight')
plt.savefig('2D_vanHove_self_correlation.eps', format='eps', dpi=400, bbox_inches='tight')

# Function to test whether the van Hove function is Gaussian
def r_to_power_n(two2d_van_hove, bins, n, t):
    """
    Calculate the nth moment of the van Hove function at a specific time step.
    
    Args:
        two2d_van_hove (ndarray): 2D array of van Hove correlation function.
        bins (ndarray): Bin edges for the van Hove function.
        n (int): Power to which the radius is raised.
        t (int): Time step index.
        
    Returns:
        float: nth moment of the van Hove function.
    """
    d_r = 1 / len(bins)
    return np.sum([two2d_van_hove[t, :] * r ** n * d_r for r in bins])

# Calculate the ratio of the fourth moment to the square of the second moment
ratio = []
for t in np.arange(0, len(two2d_van_hove), 1):
    numerator = r_to_power_n(two2d_van_hove, bins, 4, t)
    denominator = r_to_power_n(two2d_van_hove, bins, 2, t)
    ratio.append(3 / 5 * numerator / denominator)

# Plot the Gaussianity test
plt.figure(figsize=(3.25, 3.25), dpi=400)
plt.plot(np.arange(0, len(two2d_van_hove)) * 50 * 0.002, ratio, marker='o')
plt.ylabel('3/5 $^*$ < $r^4$(t)> / < $r^2$(t)>', size=10)
plt.xlabel('Simulation time, t (ps)', size=10)
plt.tick_params(axis='both', direction='in')

# Save the plot for Gaussianity test
plt.savefig('Gaussianicity_test.png', format='png', dpi=400, bbox_inches='tight')
plt.savefig('Gaussianicity_test.eps', format='eps', dpi=400, bbox_inches='tight')
