# -*- coding: utf-8 -*-
"""
Created on Mon Apr 8 07:51:56 2024

@author: vlandgraf

This script analyzes molecular dynamics trajectories to identify "strings" of
atomic transitions and compares the distribution of string lengths to a Poisson
distribution using statistical tests.
"""

# Import necessary libraries
import os
import glob
import copy
import random
import math
import pickle as pkl
import numpy as np
import matplotlib.pyplot as plt
from itertools import combinations
from scipy import stats
import openpyxl
from openpyxl import Workbook
from pathlib import Path
from pymatgen.io.vasp import Vasprun
from pymatgen.io.vasp.outputs import Poscar
from gemdat import SitesData, Trajectory, plots, transitions, simulation_metrics
from gemdat.io import load_known_material
import general_AF_structure_analysis_special as general_AF_structure_analysis

# Load trajectory data from cache
dict_of_trajectories = {72: gemdat.trajectory.Trajectory.from_cache('vasprun.xml.69adc856.cache')}

# Loop through trajectories and analyze each one
for k, trajectory in dict_of_trajectories.items():

    # Retrieve the initial structure and add all Li atoms
    s = trajectory.get_structure(0)
    s = general_AF_structure_analysis.get_structure_with_all_Li(s)
    
    # Initialize SitesData object to track Li positions
    sites = SitesData(
        structure=s,
        trajectory=trajectory,
        floating_specie='Li',
        n_parts=1
    )
    
    # Get the transition events for the Li atoms
    events = sites.transitions.events
    
    def find_chains(input_event, events, counter=1, events_list=None): 
        """
        Recursively finds chains of related events where an atom moves from one site to another.
        
        Args:
            input_event (array): Initial event to start the chain from.
            events (array): Array of all events.
            counter (int): Counter for the chain length.
            events_list (list): List to store events in the chain.
        
        Returns:
            counter (int): Length of the chain.
            events_list (list): List of events in the chain.
        """
        if events_list is None:
            events_list = []

        end_time = input_event[-1]
        start_time = input_event[-2]
        end_site = input_event[-3]
        atom = input_event[0]

        # Identify strings
        events_filtered = events[
            (events[:, 0] != atom) &
            (events[:, -2] >= start_time) &
            (events[:, -2] <= end_time) &
            (events[:, 1] == end_site)
        ]

        if len(events_filtered) > 0:
            events_list.append(events_filtered[0])
            counter += 1
            return find_chains(events_filtered[0], events, counter=counter, events_list=events_list)
        else:
            return counter, events_list
        
    def find_indices_for_events(events_list, events): 
        """
        Finds the indices of a list of events within a larger array of events.
        
        Args:
            events_list (list): List of events to find indices for.
            events (array): Array of all events.
        
        Returns:
            indices (list): List of indices corresponding to the events in events_list.
        """
        return [np.where((events == e).all(axis=1))[0][0] for e in events_list]
    
    # Analyze the first event to start the process
    string, event_list = find_chains(events[0], events, events_list=[events[0]])
    indices = find_indices_for_events(event_list, events)
    
    # Main loop: process all events to find chains
    list_of_events_with_event_indices = []
    for e in events:
        string, events_in_string_list = find_chains(e, events, events_list=[e])
        indices = find_indices_for_events(events_in_string_list, events)
        list_of_events_with_event_indices.append(indices)
    
    # Filter out double counting of events
    filtered_list_of_events_with_event_indices = []
    for event_index in range(len(events)): 
        indices = [(sublist, len(sublist)) for sublist in list_of_events_with_event_indices if event_index in sublist]
        sorted(indices, key=lambda x: x[1], reverse=True)
        filtered_list_of_events_with_event_indices.append(indices[0][0])
    
    # Analyze the distribution of string lengths
    string_length_distribution = np.array([len(i) - 1 for i in filtered_list_of_events_with_event_indices])
    number_of_strings = len(np.where(string_length_distribution > 0)[0])
    average_length_of_strings = np.sum(string_length_distribution) / len(string_length_distribution)
    
    # Perform Cramér-von Mises test to compare to Poisson distribution
    cvm_statistic = stats.cramervonmises(string_length_distribution, 'poisson', args=(average_length_of_strings,)).statistic
    
    # Generate statistics for 100 random Poisson distributions of similar length
    cvm_of_distributions = [
        stats.cramervonmises(
            np.random.poisson(average_length_of_strings, len(string_length_distribution)),
            'poisson',
            args=(average_length_of_strings,)
        ).statistic
        for _ in range(100)
    ]
    
    # Generate a larger Poisson distribution for comparison
    poisson_string_length_distribution = np.random.poisson(average_length_of_strings, 100000)
    
    # Calculate unique counts for observed and random distributions
    unique, count = np.unique(np.array(string_length_distribution), return_counts=True)
    unique_random, count_random = np.unique(np.array(poisson_string_length_distribution), return_counts=True)
    
    # Plot the distribution of string lengths
    plt.figure(figsize=(3.25, 3.25), dpi=400)
    plt.plot(unique, count / np.sum(count), c='blue', marker='o', label='AIMD')
    plt.plot(unique_random, count_random / np.sum(count_random), c='red', marker='o', label='Poisson')
    plt.yscale('log')
    plt.legend()
    plt.tick_params(axis='both', which='both', direction='in', labelsize=9)
    
    max_x = np.max([np.max(unique), np.max(unique_random)])
    x_axis = np.arange(0, max_x, 1)
    plt.xticks(x_axis, x_axis + 1)
    plt.xlim(-0.5, max_x + 0.5)
    plt.ylabel('P(n)', size=12)
    plt.xlabel('String length n', size=12)
    
    # Save the plot
    plt.savefig(f'string_length_dist_{k}.png', format='png', dpi=400, bbox_inches='tight')
    
    # Save the statistics to an Excel file
    file_path = 'statistic_of_the_distributions_of_string_lengths.xlsx'
    
    wb = openpyxl.load_workbook(file_path)
    sheet = wb.create_sheet(title=str(k))
    
    # Headers for the Excel sheet
    headers = ["CVM for Distribution", "Average CVM for Poisson_same_length", "CVM for Poisson_same_length_std"]
    sheet.append(headers)
    
    # Append the calculated statistics
    sheet.append([cvm_statistic, np.average(cvm_of_distributions), np.std(cvm_of_distributions)])
    
    # Save the workbook
    wb.save(file_path)
